/*  VT1500 Device Driver for Linux, Version 1.0.10
 *  Copyright (C) 1995-2001  M. Gutschke
 *
 *  At the time of writing my e-mail address is:
 *	Internet: markus@gutschke.com
 *  My snail mail address is:
 *	Markus Gutschke
 *	3637 Fillmore Street #106
 *	San Francisco, CA 94123-1600
 *	USA
 *  If you like this software, I would appreciate if you sent me a postcard
 *  from your hometown. Under the terms of the GNU general public license
 *  you are free to include this program into (commercial) software
 *  distributions (e.g. putting it onto CD-ROM); nonetheless, I would really
 *  appreciate if you dropped me a short note (sending me a sample copy of
 *  your distribution would be even more appreciated!)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <errno.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#include "server.h"
#include "rc.h"
#include "rcall.h"

static int readn(int fd,char *ptr,int nbytes)
{
  int nleft,nread;

  for (nleft = nbytes; nleft > 0; nleft -= nread, ptr += nread)
    if ((nread = read(fd,ptr,nleft)) < 0) {
      fprintf(stderr,"tv: Read error on socket: %s\n",strerror(errno));
      exit(1); }
    else if (nread == 0) {
      fprintf(stderr,"tv: Reached end of file on socket: %s\n",
	      strerror(errno));
      exit(1); }
  return(nbytes - nleft);
}

static int writen(int fd,const char *ptr,int nbytes)
{
  int nleft,nwritten;

  for (nleft = nbytes; nleft > 0; nleft -= nwritten, ptr += nwritten)
    if ((nwritten = write(fd,ptr,nleft)) <= 0) {
      fprintf(stderr,"tv: Write error on socket: %s\n",strerror(errno));
      exit(1); }
  return(nbytes - nleft);
}

static void discardInput(int fd)
{
  char buf[128];
  int  readable;

  for (;;) {
    if (ioctl(fd,FIONREAD,&readable)) {
      fprintf(stderr,"tv: FIONREAD on socket failed: %s\n",strerror(errno));
      exit(1); }
    if (readable > sizeof(buf))
      readable = sizeof(buf);
    else if (!readable)
      return;
    readn(fd,buf,readable); }
}

static char *readline(int fd)
{
  static char buffer[256];
  int         readable,idx = 0;

  do {
    readn(fd,buffer+idx++,1); /* Avoid busy waiting */
    for (;;) {
      if (ioctl(fd,FIONREAD,&readable)) {
	fprintf(stderr,"tv: FIONREAD on socket failed: %s\n",strerror(errno));
	exit(1); }
      if (idx + readable >= sizeof(buffer)-1) {
	fprintf(stderr,"tv: Received illegal response: Capacity exceeded\n");
	exit(1); }
      else if (readable == 0)
	break;
      idx += readn(fd,buffer+idx,readable); }
  } while (buffer[idx-1] != '\n');
  buffer[idx-1] = '\000';
  return(buffer);
}

int rcall(const char *fncname,const char *fmt,...)
{
  extern int         inetsock;
  static int         sockfd = -1;
  va_list            arg_ptr;
  enum {tVoid,tInt,tStr,tBStr,
  }                  rcType[10];
  void               *rcPtr[10];
  int                argCount = 0,rcCount = 0,i;
  char               *ptr;

  if (inetsock >= 0)
    return(0); /* running in server mode */
  if (sockfd < 0) {
    struct sockaddr *serv_addr;
    int             addr_len;

    serv_addr = (struct sockaddr *)&RCSettings.s.server;
    addr_len  = serv_addr->sa_family == AF_UNIX ? sizeof(struct sockaddr_un)
                                                : sizeof(struct sockaddr_in);
    if ((sockfd = socket(serv_addr->sa_family,SOCK_STREAM,0)) < 0) {
      fprintf(stderr,"tv: Cannot open stream socket: %s\n",
	      strerror(errno));
      exit(1); }
    if (connect(sockfd,serv_addr,addr_len) < 0) {
      fprintf(stderr,"tv: Cannot connect to server: %s\n",strerror(errno));
      exit(1); } }
  discardInput(sockfd);
  if (fncname == NULL)
    return(1);
  va_start(arg_ptr,fmt);
  if (fncname[1] == ':') {
    switch (*fncname) {
    case 'v': break;
    case 'i': *rcType = tInt; *rcPtr = va_arg(arg_ptr,void *); rcCount++;break;
    case 's': *rcType = tStr; *rcPtr = va_arg(arg_ptr,void *); rcCount++;break;
    default:
      fprintf(stderr,"tv: Internal error\n");
      exit(1); }
    fncname += 2; }
  writen(sockfd,fncname,strlen(fncname));
  writen(sockfd,"(",1);
  for (; fmt && *fmt; fmt++) {
    switch (*fmt) {
    case 'i': case 's': case '&': if (argCount++) writen(sockfd,",",1); break;
    default: break; }
    switch (*fmt) {
    case 'i': {char iBuf[16]; sprintf(iBuf,"%d",va_arg(arg_ptr,int));
              writen(sockfd,iBuf,strlen(iBuf)); break; }
    case 's': {char *ptr = va_arg(arg_ptr,char *);
              writen(sockfd,"\"",1);
              writen(sockfd,ptr,strlen(ptr));
	      writen(sockfd,"\"",1);
	      break; }
    case 'I': rcType[rcCount] = tInt;
              rcPtr[rcCount++] = va_arg(arg_ptr,void *); break;
    case 'S': rcType[rcCount] = tStr;
              rcPtr[rcCount++] = va_arg(arg_ptr,void *); break;
    case '&': switch (*++fmt) {
    case 'i': {char iBuf[16];
              rcType[rcCount] = tInt;
              sprintf(iBuf,"%d",
		      *(int *)(rcPtr[rcCount++] = va_arg(arg_ptr,void *)));
	      writen(sockfd,iBuf,strlen(iBuf)); break; }
    case 's': {char *ptr;
              rcType[rcCount] = tBStr;
	      ptr = rcPtr[rcCount++] = va_arg(arg_ptr,void *);
	      writen(sockfd,"\"",1);
	      writen(sockfd,ptr,strlen(ptr));
	      writen(sockfd,"\"",1);
	      break; }
    default:  goto internal; }
              break;
    default:
    internal: fprintf(stderr,"tv: Internal error\n");
              exit(1); } }
  va_end(arg_ptr);
  writen(sockfd,");",2);
  ptr = readline(sockfd);
  if (!rcCount) {
    if (strcmp(ptr,"OK")) {
      fprintf(stderr,"tv: Protocol failure: \"OK\" expected\n");
      exit(1); } }
  else {
    for (i = 0; i < rcCount; i++) {
      if (!ptr) {
	fprintf(stderr,"tv: Missing return code\n");
	exit(1); }
      switch (rcType[i]) {
      case tInt:  if (!isdigit(*ptr) && *ptr != '-') {
	            fprintf(stderr,"tv: Illegal return code: "
		  	           "Integer expected\n");
		    exit(1); }
                  if (rcPtr[i])
		    *(int *)(rcPtr[i]) = atoi(ptr);
	          if ((ptr = strchr(ptr,',')) != NULL) ptr++;
		  break;
      case tStr:  if (*ptr != '\"') {
	            fprintf(stderr,"tv: Illegal return code: "
                                   "String expected\n");
		    exit(1); }
                  if (rcPtr[i])
		    *(char **)(rcPtr[i]) = ptr+1;
		  if ((ptr = strchr(ptr+1,'\"')) == NULL) {
		    fprintf(stderr,"tv: Protocol failure: "
			           "Mismatched quotes\n");
		    exit(1); }
		  *ptr++ = '\000';
		  if (!*ptr) ptr = NULL;
		  else if (*ptr++ != ',') {
		    fprintf(stderr,"tv: Protocol failure: "
			           "Missing comma\n");
		    exit(1); }
	          break;
      case tBStr: {char *tmp = ptr+1;
                   if (*ptr != '\"') {
	            fprintf(stderr,"tv: Illegal return code: "
                                   "String expected\n");
		    exit(1); }
		  if ((ptr = strchr(ptr+1,'\"')) == NULL) {
		    fprintf(stderr,"tv: Protocol failure: "
			           "Mismatched quotes\n");
		    exit(1); }
		  *ptr = '\000';
                  if (rcPtr[i])
		    strcpy((char *)(rcPtr[i]),tmp);
		  if (!*++ptr) ptr = NULL;
                  break; }
      case tVoid: /* fall thru */
      default:
	fprintf(stderr,"tv: This cannot happen...\n");
	exit(1); } }
      if (ptr != NULL) {
	fprintf(stderr,"tv: Protocol failure: Too many return codes\n");
	exit(1); } } 
  return(1);
}
